from setuptools import setup

if __name__ == '__main__':
    setup(
        name='naga',
        packages=['naga'],  # this must be the same as the name above
        version='0.1.04',
        description='A powerful Python/Lisp-hybrid-monster!',
        author='James J. Tolton',
        author_email='jjtolton@gmail.com',
        url='https://github.com/hiimjayhireme/naga',
        keywords=['testing', 'logging', 'example'],
        classifiers=[],
    )
